#webapp.py - not done yet but the docstring is kind of
#a spec - 7/3/2002

"""This provides a base class for web applications.

After some study of JSP, ASP, Twisted, Webware etc
I realised our standard controllers are oversimplifying
things.  There is a 'standard model' used by all these
established frameworks which is just complex enough to
capture what goes on.

WebApp holds any 'state' for the app and handles requests.
It should be written so it will work OK in a long running
framework and be able to handle multiple concurrent requests.
It will provide a number of methods to be overridden which
allow pre- or post-processing of requests.  

WebRequest holds any info passed into the request.
For now it has an attribute 'params' holding any parameters
in a dictionary, but it could also hold a lot more stuff
abstracted from the CGI or server nvironment e.g. calling
IP address, browser details.

WebResponse handles the output. It's just a buffer you
write HTML or other data into, but you can change its
content-type from HTML to GIF or whatever, right up to
the end.  It is very important to write to a response
and not just use 'print', as (a) this lets us write
test cases which work without going through the web and
(b) if ever we switch an app to a long running server
then print statements will be a risk and require careful
synchronization.

RlWebApp is a subclass of WebApp which has some common
knowledge about things ReportLab wants to standardise in
its web applications: logging, completion marks, user
management.  It will have knowledge of preppy views, and
of QDF files as they mature.
1. we try for a method matching the action parameter
2. we try for a prep file in the views subdirectory

A major goal is that web apps should be as fully testable
as possible from scripts without needing a browser or web
server; one should be able to prepare request/response
assertions and test them in isolation.




"""
import sys, os
from rlextra.preppy import preppy
from rlextra.radxml.xmlutils import translateStandardEntities, quoteValue

def getCgiParams():
    "returns dictionary of parameters found in the cgi script"
    dictionary = {}
    import cgi
    form = cgi.FieldStorage()
    for name in form.keys():
        value = form[name].value
        dictionary[name] = quoteValue(value)
    return dictionary

def getCommandLineParams():
    "return any key=value params in command line as a dict"
    params = {}
    for arg in sys.argv:
        chunks = arg.split('=')
        if len(chunks) == 2:
            params[chunks[0]]=chunks[1]
    return params
    

class WebRequest:
    def __init__(self):
        #expand this to put in all the things potentially retrievable
        #from the request - not just the parameters dictionary
        self.url = None
        self.params = {}
        
class WebResponse:
    def __init__(self):
        self._headers = {}
        self._outputList = []
        self.setContentType('text/html') # unless overridden

    def setHeader(self, key, value):
        self._headers[key] = value

    def setContentType(self, typeCode):
        self._headers['Content-type'] = typeCode

    def getContentType(self):
        return self._headers['Content-type']


    def write(self, stuff):
        "Make it look like a file object"
        self._outputList.append(stuff)

    def getOutput(self):
        "Return the full HTTP response as a string"
        buf = []
        data = ''.join(self._outputList)
        length = len(data)

        # conventionally content-length comes after
        # content-type; for others we do not care
        # about the order

        ctype = self._headers['Content-type']
        del self._headers['Content-type']
        buf.append('Content-type: %s\nContent-length: %d\n' % (ctype, length))
        
        for (key, value) in self._headers.items():
            buf.append('%s: %s\n' % (key, value))

        buf.append('\n')
        
        buf.append(data)
        return ''.join(buf)
    

    def respond(self, device=sys.stdout):
        """What it does at the end.

        You can hook into this more easily by testing the value
        of getOutput"""
        device.write(self.getOutput())
    
    def containsText(self, text):
        "Use for writing test cases"
        found = self.getOutput().find(text)
        return (found > -1)

    def containsPattern(self, pattern):
        "Support for test cases.  Make a regex from pattern"
        pass
    
    def getHeader(self, headerType):
        "Support for test cases.  Get value of header, or raise key error"
        pass
    
class WebApp:
    """Basic dispatch mechanism for web applications.

    
    """

    def __init__(self):
        self.actionSynonyms = ['action','ERS_ACTION']
        self.actionMethodPrefix = ''
        self.viewDir = 'views'
        self.defaultView = None
        self.loginManager = None
        self.passPreppyParamsOnly = 0  #legacy mode
        self.cgiName = None
        from reportlab.rl_config import T1SearchPath
        for d in '..', '.':
            d = os.path.normpath(os.path.abspath(os.path.join(d,'fonts')))
            if d not in T1SearchPath and os.path.isdir(d):
                T1SearchPath.insert(0,d)

    def authenticate(self, request):
        "Return true if allowed to do what they want"
        return 1

    def handleCgi(self):
        "Easy entry point for CGI scripts, does everything"
        sys.stderr = sys.stdout
        try:
            req = WebRequest()
            req.params = getCgiParams()
            resp = self.handleRequest(req)
            print resp.getOutput()
        except:
            print 'Content-type: text/html'
            print
            print '<pre>'
            raise

    def handleCommandLine(self):
        "Easy entry point for command line testing"
        req = WebRequest()
        req.url = 'commandline'
        req.params = getCommandLineParams()
        resp = self.handleRequest(req)
        print resp.getOutput()

    def handleRequest(self, request):
        """Accept a WebRequest, return a WebResponse.
        Ensure every request goes through this, whether invoked
        from CGI or command line or anything else."""
        response = WebResponse()
        return self._handleRequest(request,response)

    def _handleRequest(self, request, response):
        #0. before hook
        #1. any security checking / authentication and lookup
        # of user details (LoginManager?)
        try:
            actionName = self.getActionName(request)
            if actionName is None:
                self.defaultAction(request, response)
            else:
                # first try at dispatch:  look for a method
                methodName = self.actionMethodPrefix + actionName
                viewName = self.viewDir + os.sep + actionName
                                
                method =  getattr(self, methodName, None)
                if method is not None:
                    apply(method, [request, response])
                elif os.path.isfile(viewName + '.pyc') or os.path.isfile(viewName + '.prep'):
                    # to do - handle failed lookups
                    mod = preppy.getModule(viewName)
                    # question - how to support
                    if self.passPreppyParamsOnly:
                        output = mod.getOutput(request.params)
                    else:
                        dict = {'controller': self,
                                'request':request,
                                'response':response}
                        output = mod.getOutput(dict)
                    response.write(output)
                else:
                    self.defaultAction(request, response)
            if response.getContentType() == 'text/html':
                self.applySuccessCompletionMark(request, response)
            return response

        except:
            error = None # what do I do to get the last error?
            #error got this far, handle it...
            self.handleError(request, response, error)
            self.applyErrorCompletionMark(request, response, error)
            return response

    def handleError(self, request, response, error):
        response.setContentType('text/html')
        response.write('<pre>An error occurred...')
        raise

    def getActionName(self, request):
        "Work out the name of the main action. Override if needed"
        for key in self.actionSynonyms:
            try:
                return request.params[key]
            except KeyError:
                pass
        # if no action given, default is expected
        return None            

           
    def defaultAction(self, request, response):
        "Do whatever you do if no arguments supplied...."
        if self.defaultView is not None:
            viewName = self.viewDir + os.sep + self.defaultView
            mod = preppy.getModule(viewName)
            if self.passPreppyParamsOnly:
                output = mod.getOutput(request.params)
            else:
                dict = {'controller': self,
                        'request':request,
                        'response':response}
                output = mod.getOutput(dict)
            response.write(output)
        else:            
            response.write("Server alive, no action specified!")

    def applySuccessCompletionMark(self, request, response):
        "Tiny page footer which is easy to parse"
        response.write('\n<!--COMPLETED:SUCCESSFULLY-->\n')
        
    def applyFailureCompletionMark(self, request, response, error):
        "Tiny page footer which is easy to parse"
        response.write('<!--COMPLETED:ERROR-->\n')

    def ping(self, request, response):
        "Test function always there"
        response.setContentType('text/plain')
        response.write('pong\n')

    def getUniqueFileName(self, prefix='out',suffix=None,ext='.pdf'):
        # builds one from timestamp
        if prefix and prefix[-1]!='-': prefix += '-'
        if not hasattr(self,'_fnc'): self._fnc = 0
        else: self._fnc = self._fnc+1
        if not suffix:
            import time
            from random import randint
            return '%s%0.0f%06d%d%s' % (prefix, 1000*time.time(), randint(0,1000000), self._fnc, ext)
        return prefix+suffix+str(self._fnc)+ext


if __name__=='__main__':
    "Allow a command line test with key=value parameter passing"
    WebApp().handleCommandLine()
