#copyright ReportLab Inc. 2001-2003
#see license.txt for license details
#history www.reportlab.co.uk/rl-cgi/viewcvs.cgi/rlextra/graphics/quickchart.py
#$Header: /rl_home/repository/rlextra/graphics/quickchart.py,v 1.50 2003/07/01 15:12:14 robin Exp $
__version__=''' $Id: quickchart.py,v 1.50 2003/07/01 15:12:14 robin Exp $ '''



"""
Handy constructor for a decent range of Excel-like charts.

Our current model assumes that you explicitly position
things on a drawing, and there is no automatic layout.
In some cases - particularly when embedding the chart library
in another app - there is a need for something 'quick and simple'
which makes some automatic layout decisions.

This adds a 'layer over the top' which can quickly create
a drawing to order, with title, legend, chart, various
labels and so on - and which will make vaguely sensible
decisions about the space taken up by each.

A second goal is to provide a finite and consistent
set of properties, so that anyone trying to make
(for example) a simple chart-editing GUI in their
app can do so.  If you are working with, say, the Quark
API, it is almost impossible to create a drill-down
property editor.


Anyone wishing to do further customisation than this
function permits can then set properties on the Drawing
which it returns.

The layout logic first looks at the number of
'things' in the drawing:  title, legend, chart,
axis labels.  It notes which are present or
absent, and the position of the legend.  It divides


"""
import sys
import string
from types import StringType, ListType, TupleType

from reportlab.lib.colors import CMYKColor, PCMYKColor, Color, red, black, pink, toColor
from reportlab.graphics.shapes import Drawing, String, Rect, Line, Group
from reportlab.graphics.widgetbase import Widget
from reportlab.graphics.widgets.markers import Marker
from reportlab.lib.validators import _isColor
from rlextra.graphics.layout import getBounds, Sizer

class ChartError(Exception):
    pass


ALLOWED_CHART_TYPES = ['area',
                       'bar', 'clustered_bar', 'percent_bar', 'stacked_bar',
                       'bubble',
                       'column', 'clustered_column', 'percent_column', 'stacked_column',
                       'doughnut',
                       'line', 'line_markers',
                       'pie', 'exploded_pie',
                       'radar', 'filled_radar', 'radar_markers',
                       'scatter', 'scatter_lines', 'scatter_lines_markers',
                       ]

ALLOWED_LEGEND_POSITIONS = [None,
                     'left','right','top','bottom',
                     'top_left','top_right','bottom_right','bottom_left',
                     'middle']

ALLOWED_MARKERS = [None,
                     'FilledSquare', 'FilledDiamond', 'FilledCircle', 'FilledCross',
                     'FilledTriangle', 'FilledStarSix', 'FilledPentagon', 'FilledHexagon',
                     'FilledHeptagon', 'FilledOctagon', 'Sequence']

#colour names as comments at the end of each line are as a memory jogger ONLY
#NOT HTML named colours!

#Main colours as used for bars etc
mainColours = [
        PCMYKColor(40,40,0,0),    # Lavender
        PCMYKColor(0,66,33,39),   # Maroon
        PCMYKColor(0,0,20,0),     # Yellow
        PCMYKColor(20,0,0,0),     # Cyan
        PCMYKColor(0,100,0,59),   # Purple
        PCMYKColor(0,49,49,0),    # Salmon
        PCMYKColor(100,49,0,19),  # Blue
        PCMYKColor(20,20,0,0),    # PaleLavender
        PCMYKColor(100,100,0,49), # NavyBlue
        PCMYKColor(0,100,0,0)     # Purple
        ]
#Highlight colors - eg for the tops of bars
highlightColours = [
        PCMYKColor(39,39,0,25),   # Light Lavender
        PCMYKColor(0,66,33,54),   # Light Maroon
        PCMYKColor(0,0,19,25),    # Light Yellow
        PCMYKColor(19,0,0,25),    # Light Cyan
        PCMYKColor(0,100,0,69),   # Light Purple
        PCMYKColor(0,49,49,25),   # Light Salmon
        PCMYKColor(100,49,0,39),  # Light Blue
        PCMYKColor(19,19,0,25),   # Light PaleLavender
        PCMYKColor(100,100,0,62), # Light NavyBlue
        PCMYKColor(0,100,0,25)    # Light Purple
        ]
#Lowlight colors - eg for the sides of bars
darkColours = [
        PCMYKColor(39,39,0,49),   # Dark Lavender
        PCMYKColor(0,66,33,69),   # Dark Maroon
        PCMYKColor(0,0,20,49),    # Dark Yellow
        PCMYKColor(20,0,0,49),    # Dark Cyan
        PCMYKColor(0,100,0,80),   # Dark Purple
        PCMYKColor(0,50,50,49),   # Dark Salmon
        PCMYKColor(100,50,0,59),  # Dark Blue
        PCMYKColor(20,20,0,49),   # Dark PaleLavender
        PCMYKColor(100,100,0,79), # Dark NavyBlue
        PCMYKColor(0,100,0,49)    # Dark Purple
        ]
#for standard grey backgrounds
backgroundGrey = PCMYKColor(0,0,0,24)



def getTestDrawing():
    "Helper for testing other programs.  Always returns the same chart!"
    d = Drawing(300,200)
    d.add(String(150,180,'Test Drawing with column chart and bar code',textAnchor='middle'))

    chart = getColumnChart()
    chart.x = 50
    chart.y = 65
    chart.width = 225
    chart.height = 100
    d.add(chart)

    from rlextra.graphics.barcode.widgets import BarcodeCode128
    bc = BarcodeCode128()
    bc.value = 'AB-12345678'
    bc.x = 100
    bc.y = 20
    d.add(bc)
    d.add(String(130,10,'AB-12345678',textAnchor='start',fontName='Courier',fontSize=7))
    return d


class DrawingLayout:
    "Data packet for where things go"
    pass

def _simpleData(data):
    '''if list of sequences return first else return thing itself'''
    if type(data[0]) in (ListType, TupleType): return data[0]
    return data

def _makeFakes(n,names,label):
    if names is None: return None
    elif type(names) is type(''):
        names = names.split()
        if len(names)==1:
            if names[0].lower()=='none': return None
            if names[0]=='auto':
                return tuple(map(lambda i, label=label: ('%s %d'%(label,i)), xrange(n)))
    i = len(names)
    if i>n: names = names[:n]
    elif i<n:
        names = list(names[:])+(n-i)*['']
    return tuple(names)

def quickChart(chartType='column',

               width=400,
               height=270,

               #array of data for the chart.
               data=None,
               categoryNames=None,
               seriesNames=None,

               #series arrangement
               seriesRelation=None,  #or 'stacked' or 'percent'

               #colors for chart
               chartColors = None,

               #chart title
               titleText = '',
               titleFontName = 'Helvetica',
               titleFontSize = None,
               titleFontColor = black,

               #big labels saying what each axis is
               xTitleText = '',
               xTitleFontName = 'Helvetica',
               xTitleFontSize = None,
               xTitleFontColor = black,

               yTitleText = '',
               yTitleFontName = 'Helvetica',
               yTitleFontSize = None,
               yTitleFontColor = black,

               #visibility and text properties for each axis
               xAxisVisible = 1,
               xAxisGridLines = 0,  #0 or 1
               xAxisFontName = 'Helvetica',
               xAxisFontSize = None,
               xAxisFontColor = black,

               yAxisVisible = 1,
               yAxisGridLines = 0, # 0 or 1
               yAxisFontName = 'Helvetica',
               yAxisFontSize = None,
               yAxisFontColor = black,

               #data labels.
               dataLabelsType=None,  #None, values, percentage, labels/fields(?)
               dataLabelsFontName = 'Helvetica',
               dataLabelsFontSize = None,
               dataLabelsFontColor = black,

               #markers for lines eg, in scatter_lines_markers
               markerType = None,
               markerSize = 6,

               # legend properties.
               legendPos = 'right',
               legendText = None,
               legendFontName = 'Helvetica',
               legendFontSize = None,
               legendFontColor = None,

               # background of entire drawing
               bgColor = None,
               plotColor = backgroundGrey,

               chartFillColors=None,
               chartStrokeColors=None,
               chartStrokeWidth=None,
               chartSeparation=0,

               #'advanced' arguments which may be needed by
               #a power user
               showBoundaries = 0, #show how space is divided up,

               drawingClass = Drawing
               ):

    "Returns a drawing containing chart, title, and possibly legend."
    #f=open("/debug.txt","w")
    #f.write(xAxisFontColor)
    #f.close()

    ###########################################################################
    #
    #  Phase 1 - check and normalize inputs.  Accept string input where possible
    #
    ###########################################################################
    #raise Exception('names = %s' % categoryNames)


    #check inputs as early as possible.  IN general we allow string inputs
    if chartType not in ALLOWED_CHART_TYPES:
        raise ChartError("chart type '" + chartType + \
                         '" not understood.  Allowed types are ' + \
                         str(ALLOWED_CHART_TYPES))
    assert legendPos in (None,
                         'left','right','top','bottom',
                         'top_left','top_right','bottom_right','bottom_left',
                         'middle'), "Unexpected legend position %s" % legendPos

    width = float(width)
    height = float(height)

    #sort out types:
    if seriesRelation in ['None', None]:
        seriesRelation = 'sidebyside'  #default
    # BAR subtypes
    if chartType == 'clustered_bar':
        chartType = 'bar' # subtype same as basic bar
        seriesRelation = 'sidebyside'
    elif chartType == 'stacked_bar':
        chartType = 'bar'
        seriesRelation='stacked'
    elif chartType == 'percent_bar':
        chartType = 'bar'
        seriesRelation='percent'
    # COLUMN subtypes
    elif chartType == 'clustered_column':
        chartType = 'column' # subtype same as basic column
        seriesRelation = 'sidebyside'
    elif chartType == 'stacked_column':
        chartType = 'column'
        seriesRelation='stacked'
    elif chartType == 'percent_column':
        chartType = 'column'
        seriesRelation='percent'

    if data is not None:
        if type(data) is StringType:
            data = parseDataBlock(data)
    _data = data

    oneLevel = chartType in ('pie','exploded_pie')
    if oneLevel:
        nSeries = 1
        data = _simpleData(data)
        nCategories = len(data)
    else:
        nSeries = len(data)
        nCategories = len(data[0])

    # check series and category names - fake them if required
    seriesNames = _makeFakes(nSeries,seriesNames,'series')
    categoryNames = _makeFakes(nCategories,categoryNames,'category')

    #compute font sizes to pass into chart.  There are many optional font sizes
    #and it saves us code to assign the default values early.
    FONT_SIZE_BIG = 0.075 * height
    FONT_SIZE_MEDIUM = 0.05 * height
    FONT_SIZE_SMALL = 0.0375 * height
    if titleFontSize is None:
        titleFontSize = FONT_SIZE_BIG
    if xTitleFontSize is None:
        xTitleFontSize = FONT_SIZE_MEDIUM
    if yTitleFontSize is None:
        yTitleFontSize = FONT_SIZE_MEDIUM
    if xAxisFontSize is None:
        xAxisFontSize = FONT_SIZE_SMALL
    if yAxisFontSize is None:
        yAxisFontSize = FONT_SIZE_SMALL
    if dataLabelsFontSize is None:
        dataLabelsFontSize = FONT_SIZE_SMALL
    if dataLabelsFontColor is None:
        dataLabelsFontColor = black
    if legendFontSize is None:
        legendFontSize = FONT_SIZE_SMALL
    if legendFontColor is None:
        legendFontColor = black

    #sort out chart colours
    if chartColors == None:
        chartColors = mainColours
    else:
        newcols = []
        def addcol(col):
            try:
                newcol = toColor(col)
                newcols.append(newcol)
            except:
                raise ValueError, "unknown color %s in chartColors" % col
        if type(chartColors) in [TupleType, ListType]:
            for col in chartColors:
                if type(col)==string:
                    addcol(col)
                elif isinstance(col,Color):
                    newcols.append(col)
                else:
                    raise ValueError, "unknown color %s in chartColors" % col
        else:
            listfound = 0
            try:
                #is it already a list/tuple which has been converted to a string?
                cols = eval(chartColors)
                for col in cols:
                    if type(col) == string:
                        addcol(col)
                    else:
                        coltest = _isColor()
                        if coltest.test(col):
                            newcols.append(col)
                        else:
                            raise ValueError, "unknown color %s in chartColors" % col
                listfound = 1
            except:
                pass # OK, maybe not. silently fail
            if not listfound:
                #allow comma separated lines
                if len(string.split(chartColors, ','))>1:
                    for col in string.split(chartColors, ','):
                        addcol(string.strip(col))
                #allow space separated lines
                elif len(string.split(chartColors, ' '))>1:
                    for col in string.split(chartColors, ' '):
                        if len(col)>0:
                            addcol(col)
        if newcols != []:
            chartColors = newcols

    if type(titleFontSize) is StringType:
        titleFontSize = float(titleFontSize)
    if type(xTitleFontSize) is StringType:
        xTitleFontSize = float(xTitleFontSize)
    if type(yTitleFontSize) is StringType:
        yTitleFontSize = float(yTitleFontSize)
    if type(xAxisFontSize) is StringType:
        xAxisFontSize = float(xAxisFontSize)
    if type(yAxisFontSize) is StringType:
        yAxisFontSize = float(yAxisFontSize)
    if type(legendFontSize) is StringType:
        legendFontSize = float(legendFontSize)

    if type(titleFontColor) is StringType:
        titleFontColor = toColor(titleFontColor)
    if type(xTitleFontColor) is StringType:
        xTitleFontColor = toColor(xTitleFontColor)
    if type(yTitleFontColor) is StringType:
        yTitleFontColor = toColor(yTitleFontColor)
    if type(xAxisFontColor) is StringType:
        xAxisFontColor = toColor(xAxisFontColor)
    if type(yAxisFontColor) is StringType:
        yAxisFontColor = toColor(yAxisFontColor)
    if type(legendFontColor) is StringType:
        legendFontColor = toColor(legendFontColor)


    # make one up
    if drawingClass is Drawing:
        d = drawingClass(width, height)
    else:
        d = drawingClass()

    if bgColor == 'None':
        bgColor = None
    elif type(bgColor) is StringType:
        bgColor = toColor(bgColor)
    if bgColor:
        d.add(Rect(0,0,width,height, strokeColor=None, fillColor=bgColor))

    if dataLabelsFontName in ( 'None',None):
        dataLabelsFontName = 'Helvetica'
    if type(dataLabelsFontSize) is StringType:
        dataLabelsFontSize = float(dataLabelsFontSize)
    elif dataLabelsFontSize is None:
        dataLabelsFontSize = labelFontSize
    if type(dataLabelsFontColor) is StringType:
        dataLabelsFontColor = toColor(dataLabelsFontColor)

    if dataLabelsType in ['None',None]:
        dataLabelsFormat = None
    elif dataLabelsType == 'values':
        dataLabelsFormat = makeDataLabels('%(value)s',_data,None,None,oneLevel)
        if oneLevel: dataLabelsFormat = dataLabelsFormat[0]
    elif dataLabelsType == 'percent':
        dataLabelsFormat = percentLabels(_data, oneLevel=oneLevel)
        if oneLevel: dataLabelsFormat = dataLabelsFormat[0]
    elif dataLabelsType == 'label':
        dataLabelsFormat = makeDataLabels('%(category)s',_data,None,None,oneLevel)
        if oneLevel: dataLabelsFormat = dataLabelsFormat[0]
    elif type(dataLabelsType) is StringType and dataLabelsType.find('%(')>=0:
        dataLabelsFormat = makeDataLabels(dataLabelsType,_data,categoryNames,seriesNames,oneLevel)
        if oneLevel: dataLabelsFormat = dataLabelsFormat[0]
    else:
        dataLabelsFormat = dataLabelsType

    if plotColor == 'None':
        plotColor = None
    elif plotColor == '(default)':
        plotColor = backgroundGrey
    elif type(plotColor) is StringType:
        plotColor = toColor(plotColor)
    elif plotColor == '(default)':
        plotColor = backgroundGrey
    elif type(plotColor) is StringType:
        plotColor = toColor(plotColor)


    ########################################################
    #
    #    Lay out the chart
    #
    ########################################################


    # compute the regions needed based on some heuristics

    # first step is to divide the space into 9 regions,
    # and assign a relative weight to each.  These will
    # be used to get proportions.  prefix w = weight
    # here are the defaults:
    PAD = height * 0.05 # generic padding between objects
    TOP_USED = LEFT_USED = RIGHT_USED = BOTTOM_USED = PAD

    #figure out what objects to place
    if titleText:
        title = String(0, 0,
                       titleText,
                       fontName=titleFontName,
                       fontSize=titleFontSize,
                       textAnchor='start',
                       fillColor=titleFontColor)
        (x1,y1,x2,y2) = getBounds(title)
        titleHeight = y2 - y1
        titleWidth = x2 - x1
        title.x = 0.5*(width - titleWidth)
        title.y = height - PAD - titleHeight
        TOP_USED = 2*PAD + titleHeight
        d.add(title, 'title')

    if legendPos is not None:
        leg = getLegend()
        #showBoundary = 0
        leg.fontName = legendFontName
        leg.fontSize = legendFontSize
        leg.fillColor = legendFontColor
        lc = []
        if chartType in ['pie', 'exploded_pie', 'doughnut']:
            #pies etc
            for i in xrange(nCategories):
                t = (chartColors[i%len(chartColors)], categoryNames and str(categoryNames[i]) or '')
                lc.append(t)
        else:
            #bars/colums or other muti-series charts
            for i in xrange(nSeries):
                t = (chartColors[i%len(chartColors)], seriesNames and str(seriesNames[i]) or '')
                lc.append(t)
        leg.colorNamePairs = lc
        if legendPos in ['top','bottom']:
            leg.columnMaximum = 1
        else:
            leg.columnMaximum = (height - (TOP_USED+PAD))/legendFontSize
        leg.dx = leg.dy = legendFontSize
        (x1,y1,x2,y2) = getBounds(leg)
        legendHeight = y2 - y1
        legendWidth = x2 - x1
        d.add(leg, 'legend')

        if legendPos == 'left':
            leg.x = PAD
            LEFT_USED = PAD + legendWidth + PAD
        elif legendPos in ['top_left','bottom_left']:
            leg.x = PAD
        elif legendPos == 'right':
            leg.x = width - legendWidth - PAD
            RIGHT_USED = PAD + legendWidth + PAD
        elif legendPos in ['top_right','bottom_right']:
            leg.x = width - legendWidth - PAD
        elif legendPos in ['top','middle','bottom']:
            leg.x = 0.5 * width - 0.5 * legendWidth  # left column has nothing

        if legendPos in ['top','top_left','top_right']:
            if titleText:
                leg.y = height - (2*PAD) - titleHeight # allows for title if it's there
                TOP_USED = max(TOP_USED, 3 * PAD + titleHeight + legendHeight)
            else:
                leg.y = height - PAD
                TOP_USED = max(TOP_USED, 2 * PAD + legendHeight)

        elif legendPos in ['left','middle','right']: #center in slab
            leg.y = BOTTOM_USED + 0.5* (height - TOP_USED - BOTTOM_USED + legendHeight)
        elif legendPos in ['bottom_left','bottom','bottom_right']:
            leg.y = PAD + legendHeight
            BOTTOM_USED = PAD + legendHeight + PAD

    #for debugging, show the bounds box
    if showBoundaries:
        d.add(Rect(LEFT_USED,
                   BOTTOM_USED,
                   width - RIGHT_USED - LEFT_USED,
                   height - TOP_USED - BOTTOM_USED,
                   fillColor=None,
                   strokeColor = red))

    chart = None
    if chartType == 'column':
        chart = getColumnChart(width, height,
                               seriesRelation=seriesRelation,
                               fontSize=dataLabelsFontSize,
                               data=data,
                               chartColors = chartColors,

                               categoryNames=categoryNames,
                               seriesNames=seriesNames,
                               plotColor=plotColor,

                               xAxisVisible=xAxisVisible,
                               xAxisGridLines=xAxisGridLines,
                               xAxisFontName=xAxisFontName,
                               xAxisFontSize=xAxisFontSize,
                               xAxisFontColor=xAxisFontColor,

                               yAxisVisible=yAxisVisible,
                               yAxisGridLines=yAxisGridLines,
                               yAxisFontName=yAxisFontName,
                               yAxisFontSize=yAxisFontSize,
                               yAxisFontColor=yAxisFontColor,

                               dataLabelsFormat=dataLabelsFormat,
                               dataLabelsFontName=dataLabelsFontName,
                               dataLabelsFontSize=dataLabelsFontSize,
                               dataLabelsFontColor=dataLabelsFontColor,
                               )

    elif chartType == 'bar':
        chart = getBarChart(width, height,
                               seriesRelation=seriesRelation,
                               fontSize=dataLabelsFontSize,
                               data=data,
                               chartColors = chartColors,

                               categoryNames=categoryNames,
                               seriesNames=seriesNames,
                               plotColor=plotColor,

                               xAxisVisible=xAxisVisible,
                               xAxisGridLines=xAxisGridLines,
                               xAxisFontName=xAxisFontName,
                               xAxisFontSize=xAxisFontSize,
                               xAxisFontColor=xAxisFontColor,

                               yAxisVisible=yAxisVisible,
                               yAxisGridLines=yAxisGridLines,
                               yAxisFontName=yAxisFontName,
                               yAxisFontSize=yAxisFontSize,
                               yAxisFontColor=yAxisFontColor,

                               dataLabelsFormat=dataLabelsFormat,
                               dataLabelsFontName=dataLabelsFontName,
                               dataLabelsFontSize=dataLabelsFontSize,
                               dataLabelsFontColor=dataLabelsFontColor,
                               )

    elif chartType == 'line':
        chart = getLineChart(width, height,
                               seriesRelation=seriesRelation,
                               fontSize=dataLabelsFontSize,
                               data=data,
                               chartColors = chartColors,

                               categoryNames=categoryNames,
                               seriesNames=seriesNames,
                               plotColor=plotColor,

                               xAxisVisible=xAxisVisible,
                               xAxisGridLines=xAxisGridLines,
                               xAxisFontName=xAxisFontName,
                               xAxisFontSize=xAxisFontSize,
                               xAxisFontColor=xAxisFontColor,

                               yAxisVisible=yAxisVisible,
                               yAxisGridLines=yAxisGridLines,
                               yAxisFontName=yAxisFontName,
                               yAxisFontSize=yAxisFontSize,
                               yAxisFontColor=yAxisFontColor,

                               dataLabelsFormat=dataLabelsFormat,
                               dataLabelsFontName=dataLabelsFontName,
                               dataLabelsFontSize=dataLabelsFontSize,
                               dataLabelsFontColor=dataLabelsFontColor,
                               markerType=markerType,
                               markerSize=markerSize
                               )

    elif chartType == 'line_markers':
        chart = getLineMarkersChart(width, height,
                               seriesRelation=seriesRelation,
                               fontSize=dataLabelsFontSize,
                               data=data,
                               chartColors = chartColors,

                               categoryNames=categoryNames,
                               seriesNames=seriesNames,
                               plotColor=plotColor,

                               xAxisVisible=xAxisVisible,
                               xAxisGridLines=xAxisGridLines,
                               xAxisFontName=xAxisFontName,
                               xAxisFontSize=xAxisFontSize,
                               xAxisFontColor=xAxisFontColor,

                               yAxisVisible=yAxisVisible,
                               yAxisGridLines=yAxisGridLines,
                               yAxisFontName=yAxisFontName,
                               yAxisFontSize=yAxisFontSize,
                               yAxisFontColor=yAxisFontColor,

                               dataLabelsFormat=dataLabelsFormat,
                               dataLabelsFontName=dataLabelsFontName,
                               dataLabelsFontSize=dataLabelsFontSize,
                               dataLabelsFontColor=dataLabelsFontColor,
                               markerType=markerType,
                               markerSize=markerSize
                               )


    elif chartType == 'scatter':
        chart = getScatterPlot(width, height,
                               seriesRelation=seriesRelation,
                               fontSize=dataLabelsFontSize,
                               data=data,
                               chartColors = chartColors,

                               categoryNames=categoryNames,
                               seriesNames=seriesNames,
                               plotColor=plotColor,

                               xAxisVisible=xAxisVisible,
                               xAxisGridLines=xAxisGridLines,
                               xAxisFontName=xAxisFontName,
                               xAxisFontSize=xAxisFontSize,
                               xAxisFontColor=xAxisFontColor,

                               yAxisVisible=yAxisVisible,
                               yAxisGridLines=yAxisGridLines,
                               yAxisFontName=yAxisFontName,
                               yAxisFontSize=yAxisFontSize,
                               yAxisFontColor=yAxisFontColor,

                               dataLabelsFormat=dataLabelsFormat,
                               dataLabelsFontName=dataLabelsFontName,
                               dataLabelsFontSize=dataLabelsFontSize,
                               dataLabelsFontColor=dataLabelsFontColor,
                               markerType=markerType,
                               markerSize = markerSize
                               )

    elif chartType == 'scatter_lines':
        chart = getScatterLinesPlot(width, height,
                               seriesRelation=seriesRelation,
                               fontSize=dataLabelsFontSize,
                               data=data,
                               chartColors = chartColors,

                               categoryNames=categoryNames,
                               seriesNames=seriesNames,
                               plotColor=plotColor,

                               xAxisVisible=xAxisVisible,
                               xAxisGridLines=xAxisGridLines,
                               xAxisFontName=xAxisFontName,
                               xAxisFontSize=xAxisFontSize,
                               xAxisFontColor=xAxisFontColor,

                               yAxisVisible=yAxisVisible,
                               yAxisGridLines=yAxisGridLines,
                               yAxisFontName=yAxisFontName,
                               yAxisFontSize=yAxisFontSize,
                               yAxisFontColor=yAxisFontColor,

                               dataLabelsFormat=dataLabelsFormat,
                               dataLabelsFontName=dataLabelsFontName,
                               dataLabelsFontSize=dataLabelsFontSize,
                               dataLabelsFontColor=dataLabelsFontColor,
                               markerType=markerType,
                               markerSize=markerSize
                               )

    elif chartType == 'scatter_lines_markers':
        chart = getScatterLinesMarkersPlot(width, height,
                               seriesRelation=seriesRelation,
                               fontSize=dataLabelsFontSize,
                               data=data,
                               chartColors = chartColors,

                               categoryNames=categoryNames,
                               seriesNames=seriesNames,
                               plotColor=plotColor,

                               xAxisVisible=xAxisVisible,
                               xAxisGridLines=xAxisGridLines,
                               xAxisFontName=xAxisFontName,
                               xAxisFontSize=xAxisFontSize,
                               xAxisFontColor=xAxisFontColor,

                               yAxisVisible=yAxisVisible,
                               yAxisGridLines=yAxisGridLines,
                               yAxisFontName=yAxisFontName,
                               yAxisFontSize=yAxisFontSize,
                               yAxisFontColor=yAxisFontColor,

                               dataLabelsFormat=dataLabelsFormat,
                               dataLabelsFontName=dataLabelsFontName,
                               dataLabelsFontSize=dataLabelsFontSize,
                               dataLabelsFontColor=dataLabelsFontColor,
                               markerType=markerType,
                               markerSize = markerSize
                               )

    elif chartType == 'pie':
        chart = getPieChart(width, height,
                               data=data,
                               chartColors = chartColors,
                               dataLabelsFormat=dataLabelsFormat,
                               dataLabelsFontName=dataLabelsFontName,
                               dataLabelsFontSize=dataLabelsFontSize,
                               dataLabelsFontColor=dataLabelsFontColor,
                               )

    elif chartType == 'exploded_pie':
        chart = getExplodedPieChart(width, height,
                               data=data,
                               chartColors = chartColors,
                               dataLabelsFormat=dataLabelsFormat,
                               dataLabelsFontName=dataLabelsFontName,
                               dataLabelsFontSize=dataLabelsFontSize,
                               dataLabelsFontColor=dataLabelsFontColor,
                               )

    elif chartType == 'doughnut':
        chart = getDoughnutChart(width, height,
                               data=data,
                               chartColors = chartColors,

                               categoryNames=categoryNames,
                               seriesNames=seriesNames,
                               dataLabelsFormat=dataLabelsFormat,
                               dataLabelsFontName=dataLabelsFontName,
                               dataLabelsFontSize=dataLabelsFontSize,
                               dataLabelsFontColor=dataLabelsFontColor,
                               )

    elif chartType == 'radar':
        chart = getRadarChart(width, height,
                               data=data,
                               chartColors = chartColors,

                               categoryNames=categoryNames,
                               seriesNames=seriesNames,
                               plotColor=plotColor,
                               dataLabelsFormat=dataLabelsFormat,
                               dataLabelsFontName=dataLabelsFontName,
                               dataLabelsFontSize=dataLabelsFontSize,
                               dataLabelsFontColor=dataLabelsFontColor,
                               )

    elif chartType == 'filled_radar':
        chart = getFilledRadarChart(width, height,
                               data=data,
                               chartColors = chartColors,

                               categoryNames=categoryNames,
                               seriesNames=seriesNames,
                               plotColor=plotColor,
                               dataLabelsFormat=dataLabelsFormat,
                               dataLabelsFontName=dataLabelsFontName,
                               dataLabelsFontSize=dataLabelsFontSize,
                               dataLabelsFontColor=dataLabelsFontColor
                               )

    elif chartType == 'radar_markers':
        chart = getRadarMarkersChart(width, height,
                               data=data,
                               chartColors = chartColors,

                               categoryNames=categoryNames,
                               seriesNames=seriesNames,
                               plotColor=plotColor,
                               dataLabelsFormat=dataLabelsFormat,
                               dataLabelsFontName=dataLabelsFontName,
                               dataLabelsFontSize=dataLabelsFontSize,
                               dataLabelsFontColor=dataLabelsFontColor,
                               markerType=markerType,
                               markerSize=markerSize
                               )

    if chart is None:
        chart = String(width*0.5, height*0.5, 'Not done yet: %s chart (%d x %d)' % (chartType, width, height),
                             textAnchor='middle')
    elif chartType in ['pie', 'exploded_pie',
                       'radar', 'filled_radar',
                       'radar_markers', 'doughnut']:
        #square bounds box
        # simple for the moment - just centres square bounds box in the  middle of
        # the drawing unless it would overlap the legend etc
        # no allowance for labels and other furniture
        availWidth = width - RIGHT_USED - LEFT_USED
        availHeight = height - TOP_USED - BOTTOM_USED
        side = min(availWidth, availHeight)
        chart.width = side
        chart.height = side
        (x1,y1,x2,y2) = getBounds(chart)
        expectedWidth = x2-x1
        expectedHeight = y2-y1
        sf = min(availWidth/float(expectedWidth),availHeight/float(expectedHeight),1)
        if sf<1:
            g=Group(transform=(sf,0,0,sf,LEFT_USED-sf*x1+0.5*(availWidth-sf*expectedWidth),BOTTOM_USED-sf*y1+(availHeight-sf*expectedHeight)*0.5))
            g.add(chart,'chart')
            chart = g
        else:
            chart.x = LEFT_USED+(availWidth-side)*0.5
            chart.y = BOTTOM_USED+(availHeight-side)*0.5

        #if chartWidth > chartHeight:
        #    chart.x = chart.x + 0.5*(chartWidth-side)
        #else:
        #    chart.y = chart.y + 0.5*(chartHeight-side)

##        (z1,z2,z3,z4) = (chart.x, chart.y, chart.width, chart.height)

##        find out how much the labels and ticks 'stick out'
##         on all sides.
##        (x1,y1,x2,y2) = getBounds(chart)

##        CHART_LEFT_MARGIN = chart.x - x1
##        CHART_RIGHT_MARGIN = x2 - chart.x - chart.width
##        CHART_BOTTOM_MARGIN = chart.y - y1
##        CHART_TOP_MARGIN = y2 - chart.y - chart.height

##        #shrink the plot rectangle accordingly
##        chart.x = LEFT_USED + CHART_LEFT_MARGIN
##        chart.y = BOTTOM_USED + CHART_BOTTOM_MARGIN

##        chartWidth = chart.width - CHART_LEFT_MARGIN - CHART_RIGHT_MARGIN
##        chartHeight = chart.height - CHART_TOP_MARGIN - CHART_BOTTOM_MARGIN
##        side = min(chartWidth, chartHeight)
##        chart.width = side
##        chart.height = side
##        if chartWidth > chartHeight:
##            chart.x = chart.x + 0.5*(chartWidth-side)
##        else:
##            chart.y = chart.y + 0.5*(chartHeight-side)
##
##        raise ValueError( """<pre>
##            old chart params: x = %d, y=%d, width=%d, height = %d
##            bounds = %d,%d,%d,%d
##            margins (LRTB)= %d, %d, %d, %d
##            new chart params: x = %d, y=%d, width=%d, height = %d
##            """ % (
##            z1,z2,z3,z4,
##            x1,y1,x2,y2,
##            CHART_LEFT_MARGIN,
##            CHART_RIGHT_MARGIN,
##            CHART_BOTTOM_MARGIN,
##            CHART_TOP_MARGIN,
##            chart.x, chart.y, chart.width, chart.height,
##
##            ))


        
    else: # rectangular plot
        #first we must position the x and y axis text.

        if yTitleText:
            yTitle = String(0,0,yTitleText,
                         fontName=yTitleFontName,
                         fontSize=yTitleFontSize,
                         fillColor=yTitleFontColor,
                         textAnchor='middle'
                         )
            (x1,y1,x2,y2) = getBounds(yTitle)
            yTitleWidth = x2 - x1
            yTitleHeight = y2 - y1
            LEFT_TITLE_SPACE = 0 + yTitleHeight
        else:
            LEFT_TITLE_SPACE = 0.0

        if xTitleText:
            xTitle = String(0,0,xTitleText,
                         fontName=xTitleFontName,
                         fontSize=xTitleFontSize,
                         fillColor=xTitleFontColor,
                         textAnchor='middle'
                         )
            (x1,y1,x2,y2) = getBounds(xTitle)
            xTitleWidth = x2 - x1
            xTitleHeight = y2 - y1
            BOTTOM_TITLE_SPACE = xTitleHeight
        else:
            BOTTOM_TITLE_SPACE = 0

        #position axis text
        if yTitleText:
            group = Group()
            group.transform = (0,1,-1,0,
                               LEFT_USED + yTitle.fontSize,
                               BOTTOM_USED + BOTTOM_TITLE_SPACE + 0.5*(height - BOTTOM_USED - BOTTOM_TITLE_SPACE - TOP_USED)
                               )
            group.add(yTitle)
            d.add(group)
        if xTitleText:
            xTitle.x = LEFT_USED + LEFT_TITLE_SPACE + 0.5*(width - LEFT_USED - LEFT_TITLE_SPACE - RIGHT_USED)
            xTitle.y = BOTTOM_USED
            d.add(xTitle)

        #initial guess filling the available space with the
        #plot rectangle.
        chart.x = 0
        chart.y = 0
        chart.width = width - RIGHT_USED - LEFT_USED - LEFT_TITLE_SPACE
        chart.height = height - TOP_USED - BOTTOM_USED - BOTTOM_TITLE_SPACE

        #find out how much the labels and ticks 'stick out'
        # on all sides.
        (x1,y1,x2,y2) = getBounds(chart)
        CHART_LEFT_MARGIN = -x1
        CHART_RIGHT_MARGIN = x2-chart.width
        CHART_BOTTOM_MARGIN = -y1
        CHART_TOP_MARGIN = y2-chart.height

        #shrink the plot rectangle accordingly
        chart.x = LEFT_USED + LEFT_TITLE_SPACE + CHART_LEFT_MARGIN
        chart.y = BOTTOM_USED + BOTTOM_TITLE_SPACE + CHART_BOTTOM_MARGIN
        chart.width = chart.width - CHART_LEFT_MARGIN - CHART_RIGHT_MARGIN
        chart.height = chart.height - CHART_TOP_MARGIN - CHART_BOTTOM_MARGIN

    d.add(chart,'chart')
    return d

def checkArgs(dict):
    """Verify all arguments, converting from string representation if needed.

    Raise relevant exceptions if bad argument types are given"""
    pass

###########################################################################
#
#  Data block manipulation helpers - turn 2d array into whatever needed
#
###########################################################################


def parseDataBlock(text, asText=0):
    lines = text.split('\n')
    data = []
    for line in lines:
        stripped = line.strip()
        if stripped <> '':
            words = stripped.split()
            if asText:
                data.append(words)
            else:
                data.append(map(float, words))
    # now check all same length
    if len(data):
        maxWidth = 0
        for row in data:
            if len(row) > maxWidth:
                maxWidth = len(row)
        for row in data:
            while len(row) < maxWidth:
                row.append(None)
    return data


def getLegendSize(rowCount, fontSize):
    "Quick guesstimate at legend size"
    height = 12 * rowCount + 5
    width = 20 + (15 * fontSize)
    return (width, height)


def normalizeData(data):
    "Return new data set where each group nsums to 1"
    colCount = len(data[0])
    rowCount = len(data)

    #make an empty matrix
    import pprint
    newBlock = []
    for row in range(rowCount):
        newBlock.append([0.0] * colCount)

    #fill it normalized
    for col in range(colCount):
        colTotal = 0.0
        for row in range(rowCount):
            assert data[row][col] is not None, "Cannot accept null data points in percent style charts!"
            colTotal = colTotal + data[row][col]
        for row in range(rowCount):
            newBlock[row][col] = data[row][col] / colTotal
    return newBlock

def makeDataLabels(fmt, data, categoryNames, seriesNames,oneLevel):
    dpc = fmt.find('%(percent)')>=0
    if dpc:
        ndata = normalizeData(data)
    else:
        ndata = data

    colCount = len(data[0])
    rowCount = len(data)
    if oneLevel: rowCount = min(rowCount,1)
    category = ''
    series = ''

    R = []
    for row in range(rowCount):
        if seriesNames: series = seriesNames[row]
        r = []
        R.append(r)
        r = r.append
        for col in range(colCount):
            if categoryNames: category = categoryNames[col]
            value = data[row][col]
            if dpc:
                try:
                    percent = 100*ndata[row][col]
                except:
                    percent = 0
            r(fmt % locals())
    return R

def percentLabels(data, decimals=0, oneLevel=None):
    "Return data block of strings with e.g. '67%'"
    return makeDataLabels(decimals and '%(percent)0.2f%%' or '%(percent)d%%',data,None,None,oneLevel)

def matrixToMultiSeries(data):
    "Turns tabular block into serieses with implicit x values 0,1,2..."
    colCount = len(data[0])
    newData = []
    for row in data:
        newSeries = []
        for i in range(colCount):
            newSeries.append((i, row[i]))
        newData.append(newSeries)
    return newData

def getColumnChart(width=200, height=150,
                   fontSize=6, data=None,
                   chartColors = None,

                   categoryNames=None,
                   seriesNames=None, plotColor=backgroundGrey,
                   seriesRelation='sidebyside',

                   xAxisVisible=1,
                   xAxisGridLines=0,
                   xAxisFontName='Helvetica',
                   xAxisFontSize=6,
                   xAxisFontColor=black,

                   yAxisVisible=1,
                   yAxisGridLines=1,
                   yAxisFontName='Helvetica',
                   yAxisFontSize=6,
                   yAxisFontColor=black,

                   dataLabelsFormat=None,
                   dataLabelsFontName='Helvetica',
                   dataLabelsFontSize=6,
                   dataLabelsFontColor=black,
                   ):

    #raise Exception("format=%s" % dataLabelsFormat)
    from reportlab.graphics.charts.barcharts import VerticalBarChart
    chart = VerticalBarChart()
    valueAxis = chart.valueAxis

    for i in range(10):
        chart.bars[i].fillColor = chartColors[i%len(chartColors)]
    chart.fillColor         = plotColor

    if type(dataLabelsFormat) is ListType:
        chart.barLabelFormat = 'values'
        chart.barLabelArray = dataLabelsFormat
    else:
        chart.barLabelFormat = dataLabelsFormat
    chart.barLabels.fontName              = dataLabelsFontName
    chart.barLabels.fontSize              = dataLabelsFontSize
    chart.barLabels.fillColor              = dataLabelsFontColor

    valueAxis.labelTextFormat = None
    valueAxis.visible = yAxisVisible
    chart.categoryAxis.visible = xAxisVisible

    if seriesRelation == 'sidebyside':
        chart.categoryAxis.style = 'parallel'
    elif seriesRelation == 'stacked':
        chart.categoryAxis.style = 'stacked'
    elif seriesRelation == 'percent':
        chart.categoryAxis.style = 'stacked'
        data = normalizeData(data)

    valueAxis.labels.fontName         = yAxisFontName
    valueAxis.labels.fontSize         = yAxisFontSize
    valueAxis.labels.fillColor     = yAxisFontColor
    chart.categoryAxis.labels.fontName       = xAxisFontName
    chart.categoryAxis.labels.fontSize       = xAxisFontSize
    chart.categoryAxis.labels.fillColor       = xAxisFontColor

    if seriesRelation!='percent':
        valueAxis.forceZero  = 'near'
        valueAxis.avoidBoundFrac = (0.5, 0.1)
    else:
        valueAxis.valueMin = 0
        valueAxis.valueMax = 1
        valueAxis.avoidBoundFrac = None
    valueAxis.rangeRound = 'both'

    if data is not None:
        chart.data = data
        chart.categoryAxis.categoryNames = categoryNames
    else:
        chart.data  = [(100, 150, 180), (125, 180, 200)]
        chart.categoryAxis.categoryNames = ['North', 'South', 'Central']
    chart.groupSpacing     = 15
    if yAxisGridLines:
        valueAxis.visibleGrid = 1
    else:
        valueAxis.visibleGrid = 0
    if xAxisGridLines:
        chart.categoryAxis.visibleGrid = 1
    else:
        chart.categoryAxis.visibleGrid = 0

    valueAxis.tickLeft = 3
    chart.categoryAxis.tickDown = 3

    return chart

def getBarChart(width=200, height=150,
                   fontSize=6, data=None,
                   chartColors = None,
                   categoryNames=None,
                   seriesNames=None, plotColor=backgroundGrey,
                   seriesRelation='sidebyside',

                   xAxisVisible=1,
                   xAxisGridLines=0,
                   xAxisFontName='Helvetica',
                   xAxisFontSize=6,
                   xAxisFontColor=black,

                   yAxisVisible=1,
                   yAxisGridLines=1,
                   yAxisFontName='Helvetica',
                   yAxisFontSize=6,
                   yAxisFontColor=black,

                   dataLabelsFormat=None,
                   dataLabelsFontName='Helvetica',
                   dataLabelsFontSize=6,
                   dataLabelsFontColor=black,
                   ):

    #raise Exception("format=%s" % dataLabelsFormat)
    from reportlab.graphics.charts.barcharts import HorizontalBarChart
    chart = HorizontalBarChart()
    for i in range(10):
        chart.bars[i].fillColor = chartColors[i%len(chartColors)]
    chart.fillColor         = plotColor

    if type(dataLabelsFormat) is ListType:
        chart.barLabelFormat = 'values'
        chart.barLabelArray = dataLabelsFormat
    else:
        chart.barLabelFormat = dataLabelsFormat
    chart.barLabels.fontName = dataLabelsFontName
    chart.barLabels.fontSize = dataLabelsFontSize
    chart.barLabels.fillColor = dataLabelsFontColor


    if seriesRelation == 'sidebyside':
        chart.categoryAxis.style = 'parallel'
    elif seriesRelation == 'stacked':
        chart.categoryAxis.style = 'stacked'
    elif seriesRelation == 'percent':
        chart.categoryAxis.style = 'stacked'
        data = normalizeData(data)

    chart.valueAxis.labels.fontName = xAxisFontName
    chart.valueAxis.labels.fontSize = xAxisFontSize
    chart.valueAxis.labels.fillColor = xAxisFontColor
    chart.valueAxis.visibleGrid  = not not xAxisGridLines
    chart.valueAxis.tickUp = 3
    chart.valueAxis.visible = xAxisVisible
    chart.valueAxis.labelTextFormat = None

    chart.categoryAxis.labels.fontName = yAxisFontName
    chart.categoryAxis.labels.fontSize = yAxisFontSize
    chart.categoryAxis.labels.fillColor = yAxisFontColor
    chart.categoryAxis.visibleGrid = not not yAxisGridLines
    chart.categoryAxis.tickLeft = 3
    chart.categoryAxis.visible = yAxisVisible

    if seriesRelation!='percent':
        chart.valueAxis.forceZero = 'near'
        chart.valueAxis.avoidBoundFrac = (0.5, 0.1)
    else:
        chart.valueAxis.valueMin = 0
        chart.valueAxis.valueMax = 1
        chart.valueAxis.avoidBoundFrac = None
    chart.valueAxis.rangeRound = 'both'

    if data is not None:
        chart.data = data
        chart.categoryAxis.categoryNames = categoryNames
    else:
        chart.data             = [(100, 150, 180), (125, 180, 200)]
        chart.categoryAxis.categoryNames = ['North', 'South', 'Central']
    chart.groupSpacing     = 15

    return chart


def getLineChart(width=200, height=150,
                   fontSize=6, data=None,
                   chartColors = None,
                   categoryNames=None,
                   seriesNames=None,
                   plotColor=backgroundGrey,
                   seriesRelation='sidebyside',

                   xAxisVisible=1,
                   xAxisGridLines=0,
                   xAxisFontName='Helvetica',
                   xAxisFontSize=6,
                   xAxisFontColor=black,

                   yAxisVisible=1,
                   yAxisGridLines=1,
                   yAxisFontName='Helvetica',
                   yAxisFontSize=6,
                   yAxisFontColor=black,

                   dataLabelsFormat=None,
                   dataLabelsFontName='Helvetica',
                   dataLabelsFontSize=6,
                   dataLabelsFontColor=black,
                   markerType=None,
                   markerSize=0
                   ):

    #raise Exception("format=%s" % dataLabelsFormat)
    from reportlab.graphics.charts.lineplots import LinePlot
    chart = LinePlot()
    for i in range(10):
        chart.lines[i].strokeColor = chartColors[i%len(chartColors)]
    chart.fillColor         = plotColor

    if type(dataLabelsFormat) is ListType:
        chart.lineLabelFormat = 'values'
        chart.lineLabelArray = dataLabelsFormat
    else:
        chart.lineLabelFormat = dataLabelsFormat
    chart.lineLabels.fontName              = dataLabelsFontName
    chart.lineLabels.fontSize              = dataLabelsFontSize
    chart.lineLabels.fillColor             = dataLabelsFontColor

    chart.xValueAxis.visible = xAxisVisible
    chart.yValueAxis.visible = yAxisVisible

    chart.yValueAxis.labelTextFormat = None
    chart.yValueAxis.labels.fontName       = yAxisFontName
    chart.yValueAxis.labels.fontSize       = yAxisFontSize
    chart.yValueAxis.labels.fillColor      = yAxisFontColor
    chart.xValueAxis.labelTextFormat = None
    chart.xValueAxis.labels.fontName       = xAxisFontName
    chart.xValueAxis.labels.fontSize       = xAxisFontSize
    chart.xValueAxis.labels.fillColor      = xAxisFontColor

    chart.yValueAxis.forceZero = 'near'
    chart.yValueAxis.avoidBoundFrac = (0.5, 0.1)
    chart.yValueAxis.rangeRound = 'both'

    chart.data = matrixToMultiSeries(data)
    if xAxisGridLines:
        chart.xValueAxis.visibleGrid  = 1
    else:
        chart.xValueAxis.visibleGrid  = 0
    if yAxisGridLines:
        chart.yValueAxis.visibleGrid  = 1
    else:
        chart.yValueAxis.visibleGrid  = 0

    chart.yValueAxis.tickLeft         = 3
    chart.xValueAxis.tickUp           = 3

    if markerType == None:
        chart.lines.symbol = None
    elif markerType == "Sequence":
        for i in range(0,10):
            chart.lines[i].symbol = Marker(kind=ALLOWED_MARKERS[i+1], size=markerSize)
    else:
        for i in range(0,10):
            chart.lines[i].symbol = Marker(kind=markerType, size=markerSize)

    return chart


def getLineMarkersChart(width=200, height=150,
                   fontSize=6, data=None,
                   chartColors = None,
                   categoryNames=None,
                   seriesNames=None,
                   plotColor=backgroundGrey,
                   seriesRelation='sidebyside',

                   xAxisVisible=1,
                   xAxisGridLines=0,
                   xAxisFontName='Helvetica',
                   xAxisFontSize=6,
                   xAxisFontColor=black,

                   yAxisVisible=1,
                   yAxisGridLines=1,
                   yAxisFontName='Helvetica',
                   yAxisFontSize=6,
                   yAxisFontColor=black,

                   dataLabelsFormat=None,
                   dataLabelsFontName='Helvetica',
                   dataLabelsFontSize=6,
                   dataLabelsFontColor=black,
                   markerType="FilledTriangle",
                   markerSize = 8
                   ):
    if markerType == None: markerType = "Sequence" # make sure there are markers
    if markerSize <3: markerSize = 8 # make sure they are visible

    chart = getLineChart(width=200, height=150,
                   fontSize=fontSize,
                   data=data,
                   chartColors = chartColors,
                   categoryNames=categoryNames,
                   seriesNames=None, 
                   plotColor=plotColor,
                   seriesRelation=seriesRelation,

                   xAxisVisible=xAxisVisible,
                   xAxisGridLines=xAxisGridLines,
                   xAxisFontName=xAxisFontName,
                   xAxisFontSize=xAxisFontSize,
                   xAxisFontColor=xAxisFontColor,

                   yAxisVisible=yAxisVisible,
                   yAxisGridLines=yAxisGridLines,
                   yAxisFontName=yAxisFontName,
                   yAxisFontSize=yAxisFontSize,
                   yAxisFontColor=yAxisFontColor,

                   dataLabelsFormat=dataLabelsFormat,
                   dataLabelsFontName=dataLabelsFontName,
                   dataLabelsFontSize=dataLabelsFontSize,
                   dataLabelsFontColor=dataLabelsFontColor,
                   markerType=markerType,
                   markerSize=markerSize
                   )
    return chart


def getScatterPlot(width=200, height=150,
                   fontSize=6, data=None,
                   chartColors = None,
                   categoryNames=None,
                   seriesNames=None,
                   plotColor=backgroundGrey,
                   seriesRelation='sidebyside',

                   xAxisVisible=1,
                   xAxisGridLines=0,
                   xAxisFontName='Helvetica',
                   xAxisFontSize=6,
                   xAxisFontColor=black,

                   yAxisVisible=1,
                   yAxisGridLines=1,
                   yAxisFontName='Helvetica',
                   yAxisFontSize=6,
                   yAxisFontColor=black,

                   dataLabelsFormat=None,
                   dataLabelsFontName='Helvetica',
                   dataLabelsFontSize=6,
                   dataLabelsFontColor=black,
                   markerType=None,
                   markerSize=0
                   ):

    #raise Exception("format=%s" % dataLabelsFormat)
    from reportlab.graphics.charts.lineplots import ScatterPlot
    chart = ScatterPlot()
    for i in range(10):
        chart.lines[i].strokeColor = chartColors[i%len(chartColors)]
    chart.fillColor         = plotColor

    if type(dataLabelsFormat) is ListType:
        chart.lineLabelFormat = 'values'
        chart.lineLabelArray = dataLabelsFormat
    else:
        chart.lineLabelFormat = dataLabelsFormat
    chart.lineLabels.fontName              = dataLabelsFontName
    chart.lineLabels.fontSize              = dataLabelsFontSize
    chart.lineLabels.fillColor             = dataLabelsFontColor

    chart.xValueAxis.visible = xAxisVisible
    chart.yValueAxis.visible = yAxisVisible

    chart.yValueAxis.labelTextFormat = None
    chart.yValueAxis.labels.fontName       = yAxisFontName
    chart.yValueAxis.labels.fontSize       = yAxisFontSize
    chart.yValueAxis.labels.fillColor       = yAxisFontColor
    chart.xValueAxis.labelTextFormat = None
    chart.xValueAxis.labels.fontName       = xAxisFontName
    chart.xValueAxis.labels.fontSize       = xAxisFontSize
    chart.xValueAxis.labels.fillColor       = xAxisFontColor

    chart.yValueAxis.forceZero = 'near'
    chart.yValueAxis.avoidBoundFrac = (0.5, 0.1)
    chart.yValueAxis.rangeRound = 'both'

    chart.data = matrixToMultiSeries(data)
    if xAxisGridLines:
        chart.xValueAxis.visibleGrid  = 1
    else:
        chart.xValueAxis.visibleGrid  = 0
    if yAxisGridLines:
        chart.yValueAxis.visibleGrid  = 1
    else:
        chart.yValueAxis.visibleGrid  = 0

    chart.yValueAxis.tickLeft         = 3
    chart.xValueAxis.tickUp           = 3

    if markerType == None:
        chart.lines.symbol = None
    elif markerType == "Sequence":
        for i in range(0,10):
            chart.lines[i].symbol = Marker(kind=ALLOWED_MARKERS[i+1], size=markerSize)
    else:
        for i in range(0,10):
            chart.lines[i].symbol = Marker(kind=markerType, size=markerSize)
    return chart


def getScatterLinesPlot(width=200, height=150,
                   fontSize=6, data=None,
                   chartColors = None,
                   categoryNames=None,
                   seriesNames=None,
                   plotColor=backgroundGrey,
                   seriesRelation='sidebyside',

                   xAxisVisible=1,
                   xAxisGridLines=0,
                   xAxisFontName='Helvetica',
                   xAxisFontSize=6,
                   xAxisFontColor=black,

                   yAxisVisible=1,
                   yAxisGridLines=1,
                   yAxisFontName='Helvetica',
                   yAxisFontSize=6,
                   yAxisFontColor=black,

                   dataLabelsFormat=None,
                   dataLabelsFontName='Helvetica',
                   dataLabelsFontSize=6,
                   dataLabelsFontColor=black,
                   markerType=None,
                   markerSize=0
                   ):
    chart = getScatterPlot(width=width, height=height,
                   fontSize=fontSize, data=data,
                   chartColors = chartColors,
                   categoryNames=categoryNames, seriesNames=None, plotColor=plotColor,
                   seriesRelation=seriesRelation,
                   xAxisVisible=xAxisVisible,
                   xAxisGridLines=xAxisGridLines,
                   xAxisFontName=xAxisFontName,
                   xAxisFontSize=xAxisFontSize,
                   xAxisFontColor=xAxisFontColor,
                   yAxisVisible=yAxisVisible,
                   yAxisGridLines=yAxisGridLines,
                   yAxisFontName=yAxisFontName,
                   yAxisFontSize=yAxisFontSize,
                   yAxisFontColor=yAxisFontColor,
                   dataLabelsFormat=dataLabelsFormat,
                   dataLabelsFontName=dataLabelsFontName,
                   dataLabelsFontSize=dataLabelsFontSize,
                   dataLabelsFontColor=dataLabelsFontColor,
                   markerType=markerType,
                   markerSize=markerSize
                   )
    chart.lines.strokeWidth = 1
    for i in range(0,10):
        chart.lines[i].strokeColor = chartColors[i%len(chartColors)]
    chart.joinedLines= 1
    return chart


def getScatterLinesMarkersPlot(width=200, height=150,
                   fontSize=6, data=None,
                   chartColors = None,
                   categoryNames=None, seriesNames=None, plotColor=backgroundGrey,
                   seriesRelation='sidebyside',

                   xAxisVisible=1,
                   xAxisGridLines=0,
                   xAxisFontName='Helvetica',
                   xAxisFontSize=6,
                   xAxisFontColor=black,

                   yAxisVisible=1,
                   yAxisGridLines=1,
                   yAxisFontName='Helvetica',
                   yAxisFontSize=6,
                   yAxisFontColor=black,

                   dataLabelsFormat=None,
                   dataLabelsFontName='Helvetica',
                   dataLabelsFontSize=6,
                   dataLabelsFontColor=black,
                   markerType=None,
                   markerSize=8
                   ):
    if markerType == None: markerType = "Sequence" # make sure there are markers
    if markerSize <3: markerSize = 8 # make sure they are visible

    chart = getScatterLinesPlot(width=width, height=height,
                   fontSize=fontSize, data=data,
                   chartColors = chartColors,
                   categoryNames=categoryNames, seriesNames=None, plotColor=plotColor,
                   seriesRelation=seriesRelation,
                   xAxisVisible=xAxisVisible,
                   xAxisGridLines=xAxisGridLines,
                   xAxisFontName=xAxisFontName,
                   xAxisFontSize=xAxisFontSize,
                   xAxisFontColor=xAxisFontColor,
                   yAxisVisible=yAxisVisible,
                   yAxisGridLines=yAxisGridLines,
                   yAxisFontName=yAxisFontName,
                   yAxisFontSize=yAxisFontSize,
                   yAxisFontColor=yAxisFontColor,
                   dataLabelsFormat=dataLabelsFormat,
                   dataLabelsFontName=dataLabelsFontName,
                   dataLabelsFontSize=dataLabelsFontSize,
                   dataLabelsFontColor=dataLabelsFontColor,
                   markerType=markerType,
                   markerSize=markerSize
                   )
    return chart


def getPieChart(width=100, height=100,
               data=[1,2,3,4],
               chartColors = None,
               dataLabelsFormat=None,
               dataLabelsFontName='Helvetica',
               dataLabelsFontSize=6,
               dataLabelsFontColor=black,
               ):

    from reportlab.graphics.charts.piecharts import Pie
    chart = Pie()
    for i in range(10):
        chart.slices[i].fillColor = chartColors[i%len(chartColors)]
    chart.data = _simpleData(data)
    chart.labels = dataLabelsFormat
    chart.startAngle  = -90
    chart.simpleLabels = 0
    chart.slices.labelRadius = 1.05
    return chart


def getExplodedPieChart(width=100, height=100,
               data=[1,2,3,4],
               chartColors = None,
               dataLabels=None,
               dataLabelsFormat=None,
               dataLabelsFontName='Helvetica',
               dataLabelsFontSize=6,
               dataLabelsFontColor=black,
               ):
    chart = getPieChart(width=width, height=height,
               data=data,
               chartColors = chartColors,
               dataLabelsFormat=dataLabelsFormat,
               dataLabelsFontName=dataLabelsFontName,
               dataLabelsFontSize=dataLabelsFontSize,
               dataLabelsFontColor=dataLabelsFontColor,
               )
    chart.slices.popout = 10
    return chart


def getDoughnutChart(width=100, height=100,
               data=[[1,2,3,4],[2,4,6,10]],
               chartColors = None,
               startAngle = 0,
               direction = 'clockwise',
               categoryNames=None,
               seriesNames=None,
               dataLabelsFormat=None,
               dataLabelsFontName='Helvetica',
               dataLabelsFontSize=6,
               dataLabelsFontColor=black,
               ):
    from reportlab.graphics.charts.doughnut import Doughnut
    chart = Doughnut()
    chart.data   = data
    for i in range(10):
        chart.slices[i].fillColor = chartColors[i%len(chartColors)]
    chart.labels                  = categoryNames
    if categoryNames is None:
        categoryNames = []
        for i in range(len(data[0])):
            categoryNames.append(str(i))
    chart.startAngle          = 0
    return chart


def getRadarChart(width=100, height=100,
               data=[(125, 180, 200), (100, 150, 180)],
               chartColors = None,
               plotColor=backgroundGrey,
               categoryNames=None,
               seriesNames=None,
               dataLabelsFormat=None,
               dataLabelsFontName='Helvetica',
               dataLabelsFontSize=6,
               dataLabelsFontColor=black,
               markerType=None,
               markerSize=0
                  ):
    from reportlab.graphics.charts.spider import SpiderChart
    chart = SpiderChart()
    for i in range(10):
        chart.strands[i].fillColor = None
        chart.strands[i].strokeWidth = 1
        chart.strands[i].strokeColor = chartColors[i%len(chartColors)]
        if markerType or markerSize:
            chart.strands.markers=1
            if markerType:
                chart.strands[i].markerType = markerType
            if markerSize:
                chart.strands[i].markerSize = markerSize
    chart.strands.fontName  = dataLabelsFontName
    chart.strands.fontSize  = dataLabelsFontSize
    chart.fillColor = plotColor
    chart.data  = data
    if categoryNames is None:
        categoryNames = []
        for i in range(len(data[0])):
            categoryNames.append(str(i))
    chart.labels = categoryNames
    chart.strands.labelRadius = 1.05
    return chart


def getFilledRadarChart(width=100, height=100,
               data=[(125, 180, 200), (100, 150, 180)],
               chartColors = None,
               plotColor=backgroundGrey,
               categoryNames=None,
               seriesNames=None,
               dataLabelsFormat=None,
               dataLabelsFontName='Helvetica',
               dataLabelsFontSize=6,
               dataLabelsFontColor=black,):
    chart = getRadarChart(width=width, height=height,
               data=data,
               chartColors=chartColors,
               plotColor=plotColor,
               categoryNames=categoryNames,
               seriesNames=None,
               dataLabelsFormat=dataLabelsFormat,
               dataLabelsFontName=dataLabelsFontName,
               dataLabelsFontSize=dataLabelsFontSize,
               dataLabelsFontColor=dataLabelsFontColor)
    for i in range(10):
        chart.strands[i].fillColor = chartColors[i%len(chartColors)]
    return chart

def getRadarMarkersChart(width=100, height=100,
               data=[(125, 180, 200), (100, 150, 180)],
               chartColors = None,
               plotColor=backgroundGrey,
               categoryNames=None,
               seriesNames=None,
               dataLabelsFormat=None,
               dataLabelsFontName='Helvetica',
               dataLabelsFontSize=6,
               dataLabelsFontColor=black,
               markerType="FilledTriangle",
               markerSize = 8
                         ):
    # make sure we have some markers!
    if markerType == None:
        markerType = "FilledTriangle"
    if markerSize == 0:
        markerSize = 8
    chart = getRadarChart(width=width, height=height,
               data=data,
               chartColors=chartColors,
               plotColor=plotColor,
               categoryNames=categoryNames,
               seriesNames=None,
               dataLabelsFormat=dataLabelsFormat,
               dataLabelsFontName=dataLabelsFontName,
               dataLabelsFontSize=dataLabelsFontSize,
               dataLabelsFontColor=dataLabelsFontColor,
               markerType=markerType,
               markerSize=markerSize
                         )
    for i in range(10):
        chart.strands[i].fillColor = None
        chart.strands[i].strokeColor = chartColors[i%len(chartColors)]
        chart.strands[i].strokeWidth = 1
    chart.strands.markers = 1
    return chart


def getLegend():
    # needs some smarts.  How to do this?
    from reportlab.graphics.charts.legends import Legend
    legend = Legend()
    legend.colorNamePairs = [(mainColours[0], 'Widgets'), (mainColours[1], 'Sprockets')]
    legend.fontName       = 'Helvetica'
    legend.fontSize       = 7
    legend.dxTextSpace    = 5
    legend.dy             = 5
    legend.dx             = 5
    legend.deltay         = 5
    legend.alignment      ='right'
    return legend

from reportlab.lib.validators import    isNumber, isNumberOrNone, isListOfStringsOrNone, isListOfNumbers, \
                                        isListOfNumbersOrNone, isColorOrNone, OneOf, isBoolean, SequenceOf, \
                                        isString, EitherOr, isColorOrNone, NoneOr
from reportlab.lib.attrmap import *
class QuickChart(Widget):
    _attrMap = AttrMap(
        chartType = AttrMapValue(apply(OneOf,tuple(ALLOWED_CHART_TYPES)),"chart type"),
        width=AttrMapValue(isNumber,"Width of the quickchart"),
        height=AttrMapValue(isNumber,"Height of the quickchart"),
        data=AttrMapValue(NoneOr(EitherOr((SequenceOf(isListOfNumbersOrNone,emptyOK=0),isString))),"Data for the quickchart"),
        categoryNames=AttrMapValue(None,""),
        seriesNames=AttrMapValue(None,""),
        seriesRelation=AttrMapValue(None,""),
        chartColors =AttrMapValue(isColorOrNone,""),
        titleText =AttrMapValue(None,""),
        titleFontName =AttrMapValue(None,""),
        titleFontSize =AttrMapValue(None,""),
        titleFontColor =AttrMapValue(isColorOrNone,""),
        xTitleText =AttrMapValue(None,""),
        xTitleFontName =AttrMapValue(None,""),
        xTitleFontSize =AttrMapValue(None,""),
        xTitleFontColor =AttrMapValue(isColorOrNone,""),
        yTitleText =AttrMapValue(None,""),
        yTitleFontName =AttrMapValue(None,""),
        yTitleFontSize =AttrMapValue(None,""),
        yTitleFontColor =AttrMapValue(isColorOrNone,""),
        xAxisVisible =AttrMapValue(isBoolean,""),
        xAxisGridLines=AttrMapValue(isBoolean,""),
        xAxisFontName=AttrMapValue(None,""),
        xAxisFontSize=AttrMapValue(None,""),
        xAxisFontColor=AttrMapValue(isColorOrNone,""),
        yAxisVisible=AttrMapValue(isBoolean,""),
        yAxisGridLines=AttrMapValue(isBoolean,""),
        yAxisFontName=AttrMapValue(None,""),
        yAxisFontSize=AttrMapValue(None,""),
        yAxisFontColor=AttrMapValue(isColorOrNone,""),
        dataLabelsType=AttrMapValue(None,""),
        dataLabelsFontName =AttrMapValue(None,""),
        dataLabelsFontSize =AttrMapValue(None,""),
        dataLabelsFontColor =AttrMapValue(isColorOrNone,""),
        markerType =AttrMapValue(None,""),
        markerSize =AttrMapValue(None,""),
        legendPos =AttrMapValue(None,""),
        legendText =AttrMapValue(None,""),
        legendFontName =AttrMapValue(None,""),
        legendFontSize =AttrMapValue(None,""),
        legendFontColor =AttrMapValue(isColorOrNone,""),
        bgColor =AttrMapValue(None,""),
        plotColor =AttrMapValue(None,""),
        chartFillColors=AttrMapValue(None,""),
        chartStrokeColors=AttrMapValue(None,""),
        chartStrokeWidth=AttrMapValue(None,""),
        chartSeparation=AttrMapValue(None,""),
        showBoundaries = AttrMapValue(isBoolean,""),
        drawingClass = AttrMapValue(None,'The drawing class to use'),
        x = AttrMapValue(isNumber, desc='x offset.'),
        y = AttrMapValue(isNumber, desc='y offset.'),
        )

    def __init__(self):
        self.chartType='column'
        self.width=400
        self.height=270
        self.data=[[100,120,140,160],[110,130,150,180]]
        self.categoryNames=None
        self.seriesNames=None
        self.seriesRelation=None
        self.chartColors = None
        self.titleText = ''
        self.titleFontName = 'Helvetica'
        self.titleFontSize = None
        self.titleFontColor = black
        self.xTitleText = ''
        self.xTitleFontName = 'Helvetica'
        self.xTitleFontSize = None
        self.xTitleFontColor = black
        self.yTitleText = ''
        self.yTitleFontName = 'Helvetica'
        self.yTitleFontSize = None
        self.yTitleFontColor = black
        self.xAxisVisible = 1
        self.xAxisGridLines=0
        self.xAxisFontName='Helvetica'
        self.xAxisFontSize=None
        self.xAxisFontColor=black
        self.yAxisVisible=1
        self.yAxisGridLines=0
        self.yAxisFontName='Helvetica'
        self.yAxisFontSize=None
        self.yAxisFontColor=black
        self.dataLabelsType=None
        self.dataLabelsFontName = 'Helvetica'
        self.dataLabelsFontSize = None
        self.dataLabelsFontColor = black
        self.markerType = None
        self.markerSize = 6
        self.legendPos = 'right'
        self.legendText = None
        self.legendFontName = 'Helvetica'
        self.legendFontSize = None
        self.legendFontColor = None
        self.bgColor = None
        self.plotColor = backgroundGrey
        self.chartFillColors=None
        self.chartStrokeColors=None
        self.chartStrokeWidth=None
        self.chartSeparation=0
        self.showBoundaries = 0
        self.drawingClass = Group
        self.x = 0
        self.y = 0

    def draw(self):
        g = quickChart(
            chartType = self.chartType,
            width = self.width,
            height = self.height,
            data = self.data,
            categoryNames = self.categoryNames,
            seriesNames = self.seriesNames,
            seriesRelation = self.seriesRelation,
            chartColors = self.chartColors,
            titleText = self.titleText,
            titleFontName = self.titleFontName,
            titleFontSize = self.titleFontSize,
            titleFontColor = self.titleFontColor,
            xTitleText = self.xTitleText,
            xTitleFontName = self.xTitleFontName,
            xTitleFontSize = self.xTitleFontSize,
            xTitleFontColor = self.xTitleFontColor,
            yTitleText = self.yTitleText,
            yTitleFontName = self.yTitleFontName,
            yTitleFontSize = self.yTitleFontSize,
            yTitleFontColor = self.yTitleFontColor,
            xAxisVisible = self.xAxisVisible,
            xAxisGridLines = self.xAxisGridLines,
            xAxisFontName = self.xAxisFontName,
            xAxisFontSize = self.xAxisFontSize,
            xAxisFontColor = self.xAxisFontColor,
            yAxisVisible = self.yAxisVisible,
            yAxisGridLines = self.yAxisGridLines,
            yAxisFontName = self.yAxisFontName,
            yAxisFontSize = self.yAxisFontSize,
            yAxisFontColor = self.yAxisFontColor,
            dataLabelsType = self.dataLabelsType,
            dataLabelsFontName = self.dataLabelsFontName,
            dataLabelsFontSize = self.dataLabelsFontSize,
            dataLabelsFontColor = self.dataLabelsFontColor,
            markerType = self.markerType,
            markerSize = self.markerSize,
            legendPos = self.legendPos,
            legendText = self.legendText,
            legendFontName = self.legendFontName,
            legendFontSize = self.legendFontSize,
            legendFontColor = self.legendFontColor,
            bgColor = self.bgColor,
            plotColor = self.plotColor,
            chartFillColors = self.chartFillColors,
            chartStrokeColors = self.chartStrokeColors,
            chartStrokeWidth = self.chartStrokeWidth,
            chartSeparation = self.chartSeparation,
            showBoundaries = self.showBoundaries,
            drawingClass = self.drawingClass,
            )
        g.transform = [1,0,0,1,self.x,self.y]
        g.__dict__['width'] = self.width
        g.__dict__['height'] = self.height
        return g

def testCommandLine(args):
    usage = """quickchart.py - high level interface for easy charts.
    usage 1: runs all tests
      quickchart.py -test

    usage 2:
      quickchart.py chartType width height [option1=value1] [option2=value2] ..
    Options are the arguments to the quickchart function.  The resulting
    chart will be stored in quickchart_output.pdf.

    """
    kwargs = {}
    positionals = []
    for arg in args:
        stuff = arg.split('=')
        if len(stuff) == 2:
            key, value = stuff
            kwargs[key] = value
            args.remove(arg)
        else:
            positionals.append(arg)
    if len(positionals) <> 3:
        print usage
        return
    chartType = args[0]
    width = float(args[1])
    height = float(args[2])

    data = [[1,2,3,4],[1,4,9,16]]
    print 'type=%s, width=%0.2f, height=%0.2f' % (chartType, width, height)

    d = quickChart(chartType, width, height, data, **kwargs)
    open('quickchart_custom.pdf','wb').write(d.asString('pdf'))
    print 'saved quickchart_custom.pdf'




def testAll():
    data = [[1,2,3,4],[1,4,9,16]]
    for chartType in ALLOWED_CHART_TYPES:
        d = quickChart(chartType, 300, 200, data)
    leg = getLegend()
    print 'created %d chart types OK' % len(ALLOWED_CHART_TYPES)
    d = getTestDrawing()
    for format in ['pdf','eps','gif','png','pict']:
        stuff = d.asString(format)
        open('quickchart_test.%s' % format,'wb').write(stuff)
        print 'saved quickchart_test.%s' % format

if __name__=='__main__':
    if '-test' in sys.argv:
        testAll()
    else:
        testCommandLine(sys.argv[1:])
